/**
 * Copyright (c) 2009 Justin Aquadro
 *
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use,
 * copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following
 * conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES
 * OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 * HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY,
 * WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR
 * OTHER DEALINGS IN THE SOFTWARE.
 */

#include <SFML/System.hpp>
#include <SFML/Window.hpp>
#include <SFML/Graphics.hpp>
#include <iostream>
#include <stdexcept>
#include "system/InputController.h"
#include "system/WindowController.h"
#include "system/FontManager.h"
#include "GLView.h"

int main (int argc, char** argv)
{
	AppState& state = AppState::getState();
	WindowController& wc = WindowController::getController();
	InputController& ic = InputController::getController();
	FontManager& fontman = FontManager::getManager();

	int appInitWidth = 800;
	int appInitHeight = 600;

	sf::FloatRect windowRect(0.f, 0.f, 800.f, 600.f);
	sf::View windowView(windowRect);

	sf::WindowSettings settings;
	settings.DepthBits = 24;
	settings.StencilBits = 8;
	settings.AntialiasingLevel = 2;

	wc.window.Create(sf::VideoMode(appInitWidth, appInitHeight, 32), "SFML Window", sf::Style::Resize | sf::Style::Close, settings);
	wc.window.SetView(windowView);
	wc.window.UseVerticalSync(true);
	wc.window.SetFramerateLimit(60);
	wc.window.PreserveOpenGLStates(true);

	GLView view = GLView(appInitWidth, appInitHeight);

	if (argc <= 1) {
		std::cout << "No model file specified" << std::endl;
		return 1;
	}

	std::cout << "File: " << argv[1] << std::endl;

	try {
		view.setModelFile(std::string(argv[1]));
		view.init();

		std::string programDir = pathname(std::string(argv[0]));

		fontman.addFont("courier_b", programDir + "/resources/NotCourier-sans-Bold.ttf", 12);
	}
	catch (std::exception e) {
		std::cout << std::endl;
		std::cout << "Uncaught Initialization Exception:" << std::endl;
		std::cout << e.what() << std::endl;
		std::cout << "Press any key to continue..." << std::flush;

		std::cin.ignore( std::numeric_limits <std::streamsize> ::max(), '\n' );
		exit(1);
	}

	try {
		while (wc.window.IsOpened()) {
			ic.input.resetState();

			wc.window.Clear(sf::Color(96, 96, 96));

			sf::Event event;
			while (wc.window.GetEvent(event))
			{
				ic.input.processEvent(event);

				if (event.Type == sf::Event::Closed) {
					wc.window.Close();
				}

				if (event.Type == sf::Event::Resized) {
					view.resize(event.Size.Width, event.Size.Height);
				}
			}

			view.eventLoop();
			view.display();

			wc.window.Display();

			state.frameCounter++;
		}
	}
	catch (std::exception e) {
		std::cout << std::endl;
		std::cout << "Uncaught Runtime Exception:" << std::endl;
		std::cout << e.what() << std::endl << std::endl;
		std::cout << "Press any key to continue..." << std::flush;

		std::cin.ignore( std::numeric_limits <std::streamsize> ::max(), '\n' );
		exit(1);
	}

    return 0;
}
